@@ -1,4 +1,4 @@
-# .. Copyright (C) 2012-2020 Bryan A. Jones.
+# .. Copyright (C) 2012-2022 Bryan A. Jones.
 #
 #   This file is part of the CodeChat System.
 #
@@ -35,7 +35,7 @@ import sys
 import subprocess
 import threading
 from time import sleep
-from typing import List, Sequence
+from typing import List, Optional, Sequence
 
 # Third-party imports
 # -------------------
@@ -49,6 +49,7 @@ from watchdog.events import FileSystemEvent, PatternMatchingEventHandler
 
 # Local application imports
 # -------------------------
+from . import __version__
 from .constants import LOCALHOST, THRIFT_PORT
 from .gen_py.CodeChat_Services import EditorPlugin
 from .gen_py.CodeChat_Services.ttypes import (
@@ -93,12 +94,12 @@ class WatcherClient:
         try:
             ret = self.thrift_client.get_client(CodeChatClientLocation.browser)
         except ConnectionResetError as e:
-            print(f"Unable to connect to the CodeChat Server: {e}.", file=sys.stderr)
+            typer.echo(f"Unable to connect to the CodeChat Server: {e}.", err=True)
             sys.exit(1)
         if ret.error:
-            print(
+            typer.echo(
                 f"Error connecting to the CodeChat Server: {ret.error}.",
-                file=sys.stderr,
+                err=True,
             )
             sys.exit(1)
         assert ret.html == ""
@@ -112,15 +113,14 @@ class WatcherClient:
             # See the `docs <https://watchdog.readthedocs.io/en/latest/api.html#watchdog.observers.api.BaseObserver.schedule>`__.
             self.observer.schedule(self.event_handler, pathname, recursive=True)
         self.observer.start()
-        print(
+        typer.echo(
             f"Watcher started, monitoring the path(s) {directories} containing patterns {patterns} but ignoring {ignore_patterns}.",
-            file=sys.stderr,
         )
 
     # See the `docs <https://watchdog.readthedocs.io/en/latest/api.html#watchdog.events.FileSystemEventHandler.on_any_event>`__.
-    def on_any_event(self, event: FileSystemEvent):
+    def on_any_event(self, event: FileSystemEvent) -> None:
         if not event.is_directory:
-            print(event, event.src_path, file=sys.stderr)
+            typer.echo(event, event.src_path, err=True)
             src_path = Path(event.src_path).absolute()
             with open(src_path, encoding="utf-8", errors="backslashreplace") as f:
                 try:
@@ -128,20 +128,20 @@ class WatcherClient:
                         f.read(), str(src_path), self.client_id, False
                     )
                 except TTransport.TTransportException as e:
-                    print(
+                    typer.echo(
                         f"Unable to communicate with the CodeChat Server when requesting a render: {e}.",
-                        file=sys.stderr,
+                        err=True,
                     )
                     self.running = False
                 else:
                     if ret:
-                        print(
+                        typer.echo(
                             f"Error requesting a render from the CodeChat Server: {ret}.",
-                            file=sys.stderr,
+                            err=True,
                         )
                         self.running = False
 
-    def run(self):
+    def run(self) -> None:
         self.running = True
         try:
             while self.running:
@@ -150,51 +150,73 @@ class WatcherClient:
             pass
         self.shutdown()
 
-    def shutdown(self):
-        print("Watcher shutting down...", file=sys.stderr)
+    def shutdown(self) -> None:
+        typer.echo("Watcher shutting down...")
         try:
             ret = self.thrift_client.stop_client(self.client_id)
         except TTransport.TTransportException as e:
-            print(
-                f"Unable to communicate when the CodeChat Server when shutting down the CodeChat Client: {e}."
+            typer.echo(
+                f"Unable to communicate when the CodeChat Server when shutting down the CodeChat Client: {e}.",
+                err=True,
             )
         else:
             if ret:
-                print(f"Error shutting down the CodeChat Client: {ret}.")
+                typer.echo(f"Error shutting down the CodeChat Client: {ret}.", err=True)
         self.observer.stop()
         self.observer.join()
-        print("Watcher shut down.", file=sys.stderr)
+        typer.echo("Watcher shut down.")
 
 
+# .. _CLI interface:
+#
 # CLI interface
 # =============
 app = typer.Typer()
 
 
+# Add a ``--version`` optional globally.
+def version_callback(value: bool) -> None:
+    if value:
+        typer.echo(f"The CodeChat Server, v.{__version__}")
+        sys.exit(0)
+
+
+@app.callback()
+def common(
+    version: Optional[bool] = typer.Option(
+        None, "--version", callback=version_callback, is_eager=True
+    ),
+) -> None:
+    pass
+
+
+INSECURE_HELP = "Setting this True allows the webserver and websocket to accept connections from any address (they bind to 0.0.0.0). This is extremely dangerous and insecure. If enabled, these ports should be firewalled and carefully protected. This mode is offered to support remote development using Visual Studio Code, which provides a secure (SSH-based) connection. The default value of False allows only connections from the machine this program runs on."
+
+
 @app.command()
 def start(
-    coverage: bool = typer.Option(False, help="Run with code coverage enabled.")
+    insecure: bool = typer.Option(False, help=INSECURE_HELP),
+    coverage: bool = typer.Option(False, help="Run with code coverage enabled."),
 ) -> None:
     "Start the server."
 
-    sys.exit(_start(coverage))
+    sys.exit(_start(insecure, coverage))
 
 
 # Typer's default arguments don't work when called directly from Python. This avoid provides a function which returns a value (0 for success, a non-zero value for failure), instead of calling ``sys.exit()``. The following CLI functions follow this pattern as well.
-def _start(coverage=False) -> int:
-    print(
+def _start(insecure: bool = False, coverage: bool = False) -> int:
+    typer.echo(
         "Starting the server -- searching for an already-running instance...",
-        file=sys.stderr,
     )
 
     # Try pinging the server to see if it's up.
     try:
         client = get_client()
         assert client.ping() == ""
-        print("A running CodeChat Server instance was found.", file=sys.stderr)
+        typer.echo("A running CodeChat Server instance was found.")
         return 0
     except Exception:
-        print("No running CodeChat Server instances found.", file=sys.stderr)
+        typer.echo("No running CodeChat Server instances found.")
 
     # The server isn't up or has crashed. Stop any existing instances in case it crashed.
     ret = _stop()
@@ -203,8 +225,9 @@ def _start(coverage=False) -> int:
 
     # Start the server, now that any hung instances are terminated.
     cov_args = ["-m", "coverage", "run"] if coverage else []
+    args = ["--insecure"] if insecure else []
     p = subprocess.Popen(
-        [sys.executable, *cov_args, "-m", "CodeChat_Server", "serve"],
+        [sys.executable, *cov_args, "-m", "CodeChat_Server", "serve", "--quiet", *args],
         stdout=subprocess.PIPE,
         stderr=subprocess.PIPE,
         text=True,
@@ -217,9 +240,9 @@ def _start(coverage=False) -> int:
         sleep(10)
         if server_starting:
             p.terminate()
-            print(
+            typer.echo(
                 "The server failed to start before the timeout expired.",
-                file=sys.stderr,
+                err=True,
             )
 
     # Make this a daemon thread, so it will exit when the main thread finishes.
@@ -227,7 +250,7 @@ def _start(coverage=False) -> int:
     timeout_thread.start()
 
     # Wait for the server to start.
-    print("Waiting for the server to start...", file=sys.stderr)
+    typer.echo("Waiting for the server to start...")
     out = ""
     line = ""
     while "CODECHAT_READY\n" not in line:
@@ -236,15 +259,15 @@ def _start(coverage=False) -> int:
         p.stderr.flush()
         line = p.stderr.readline()
         out += line
-        print(line, end="")
+        typer.echo(line, nl=False)
         if p.poll() is not None:
             # The server shut down. Print the output collected to help the user understand what went wrong.
             #
             # Make mypy happy.
             assert p.stdout is not None
-            print(p.stdout.read())
-            print(p.stderr.read())
-            print("The server exited unexpectedly.", file=sys.stderr)
+            typer.echo(p.stdout.read())
+            typer.echo(p.stderr.read(), err=True)
+            typer.echo("The server exited unexpectedly.", err=True)
             return 1
         sleep(0.001)
     # Indicate the server started to the timeout thread, so the timeout won't kill the server now that it's started.
@@ -259,7 +282,7 @@ def stop() -> None:
 
 
 def _stop() -> int:
-    print("Stopping all CodeChat Server instances...", file=sys.stderr)
+    typer.echo("Stopping all CodeChat Server instances...")
     # Look for the server. TODO: should I avoid hard-coding this?
     server_name = "CodeChat_Server"
     ret = 0
@@ -279,29 +302,33 @@ def _stop() -> int:
             # If we have the command line, then "serve" must be in it.
             and (not p.info["cmdline"] or "serve" in p.info["cmdline"])
         ):
-            print(
+            typer.echo(
                 f"Killing server process {p.pid} named {p.info['name']} with command line {p.info['cmdline']}.",
-                file=sys.stderr,
             )
             # Killing the parent of a CodeChat Server process will kill the child; ignore the exception in this case.
             try:
                 p.kill()
             except psutil.NoSuchProcess:
-                # Report an error, but continue trying to kill other processes first.
+                pass
+            except Exception as e:
+                typer.echo(f"Unable to kill: {e}.", err=True)
                 ret = 1
 
     return ret
 
 
 @app.command()
-def serve() -> None:
+def serve(
+    insecure: bool = typer.Option(False, help=INSECURE_HELP),
+    quiet: bool = typer.Option(False, help="Omit diagnostic messages."),
+) -> None:
     "Run the server in the current terminal/console."
 
     # This file takes a long time to load and run. Print a status message as it starts.
-    print("Loading...", file=sys.stderr)
+    typer.echo("Loading...")
     from .server import run_servers
 
-    sys.exit(run_servers())
+    sys.exit(run_servers(insecure, quiet))
 
 
 @app.command()
@@ -311,12 +338,13 @@ def build(path_to_build: List[Path]) -> None:
     from .renderer import render_file
 
     async def aprint(_str):
-        print(_str, end="")
+        typer.echo(_str, err=True, nl=False)
 
     # Build each path provided.
     for ptb in path_to_build:
         ptb = ptb.resolve()
-        print(f"Building {ptb}...", file=sys.stderr)
+        # In this function, send ALL status messages to stderr, so that a build can produce only the build output to stdout.
+        typer.echo(f"Building {ptb}...", err=True)
         (
             was_performed,
             rendered_project_path,
@@ -326,20 +354,20 @@ def build(path_to_build: List[Path]) -> None:
         ) = asyncio.run(render_file(file_text(ptb), str(ptb), None, aprint, False))
         assert was_performed
         # Print all errors produced by the render.
-        print(err_string, file=sys.stderr)
+        typer.echo(err_string, err=True)
         # Dump the HTML produced.
         if html is None:
-            print(
+            typer.echo(
                 f"The rendered result is stored in {rendered_file_path}.",
-                file=sys.stderr,
+                err=True,
             )
         else:
             if ptb.is_file():
-                print(html)
+                typer.echo(html)
             else:
-                print(
+                typer.echo(
                     f"Error: file {ptb} not found, and no containing project to render was found.",
-                    file=sys.stderr,
+                    err=True,
                 )
                 sys.exit(1)
 
@@ -348,7 +376,7 @@ def build(path_to_build: List[Path]) -> None:
 def render(path_to_build: Path, id: int) -> None:
     "Render the specified CodeChat project in a web browser."
 
-    print(f"Rendering {path_to_build} using ID {id}.", file=sys.stderr)
+    typer.echo(f"Rendering {path_to_build} using ID {id}.")
     ret = _start()
     if ret:
         sys.exit(ret)
